! io.f90: input/output module
!
! Author: Yang Liu
!
! Last modified: 08/18/2014
!
! Dependency:
!   test: configuration of a test

module io
use test
implicit none
public

! input/output public variables
!   fstem: file stem
!   cfile: control file
!   yfile: data file
!   pfile: parameter starting values
!   zfile: latent variable starting values
!   fscore: whether to output factor score

character(len = 80) :: cfile, yfile, pfile, zfile
character(len = 80) :: fstem
logical :: fscore = .false.

! temporary (public) variables for I/O
!   ncat: number of response categories
!   nfixed: number of fixed parameters
!   which: indices for fixed parameters (starts from 0)
!   values: fixed values
!   fscore: whether to output factor score
integer :: ncat, nfixed
integer, allocatable :: fixed(:)
double precision, allocatable :: at_val(:), start_val(:)

! namelist setup
!   files: data file, parameter and latent variable starting values
!   testspec: test specification
!   mcmcspec: MCMC sampling setup
!   itemspec: item specification (must be input once for each item)
!   itemconstr: parameter constraints
!   output: output setup
namelist /files/ yfile, pfile, zfile
namelist /testspec/ nobsn, nitem, nfac
namelist /mcmcspec/ seed, max_iter, burnin, thin, param_bnd
namelist /itemspec/ ncat, nfixed
namelist /itemparam/ start_val, fixed, at_val
namelist /output/ fstem, fscore

contains

! input: Input test configuration
!
! Arguments: none
!
! Effect: assign global variables

subroutine input_nml()
  integer :: j

  ! Open control file and read information. Comments are allowed at the end of
  ! each line
  open(100, file = cfile, status = "old", action = "read")
  read(100, nml = files)
  read(100, nml = testspec)
  read(100, nml = mcmcspec)
  call inittest()  ! initialize test
  call readdata()  ! read data files
  call readzstart()  ! read starting z values
  do j = 1, nitem
    read(100, nml = itemspec)
    items(j)%ncat = ncat
    items(j)%npar = ncat + nfac + 1
    call inititem(j)  ! initialize item j
    allocate( fixed(nfixed), at_val(nfixed), start_val(items(j)%npar) )
    read(100, nml = itemparam)
    items(j)%par = start_val  ! starting values for all parameters
    items(j)%fi(fixed + 1) = fixed  ! indices of fixed parameters
    items(j)%par(fixed) = at_val  ! fixed at values
    call initpoly(j)  ! initialize polytope for item j
    deallocate(fixed, at_val, start_val)
  end do
  read(100, nml = output)
  close(100)

end subroutine input_nml

! readdata: Read data file
! readzstart: Read starting values for z
!
! Arguments: none
!
! Effect: Data and starting values input

subroutine readdata()
  integer :: i, j
  open(101, file = yfile, status = "old", action = "read")
  do i = 1, nobsn
    read(101, *) (y(i, j), j = 1, nitem)
  end do
  close(101)
end subroutine readdata

subroutine readzstart()
  integer :: i, d
  open(103, file = zfile, status = "old", action = "read")
  do i = 1, nobsn
    read(103, *) (z(i, d), d = 1, nfac)
  end do
  close(103)
end subroutine readzstart

subroutine input()
  integer :: rstat
  integer :: i, j, p
  double precision :: val
  character(len = 40) :: cstr
  character(len = 2) :: key
  !integer :: dot

  ! Extract file stem from cfile by finding the position of '.' and discard the
  ! extension name of cfile
  !dot = index(cfile, '.', .false.)
  !fstem = cfile(1:dot - 1);

  ! Open control file and read information. Comments are allowed at the end of
  ! each line
  open(100, file = cfile, status = "old", action = "read")
  read(100, *) yfile, pfile, zfile
  read(100, *) nobsn, nitem
  read(100, *) nfac
  read(100, *) seed, max_iter, burnin, param_bnd

  ! Read data
  allocate( y(nobsn, nitem) )
  open(101, file = yfile, status = "old", action = "read")
  do i = 1, nobsn
    read(101, *) (y(i, j), j = 1, nitem)
  end do
  close(101)

  ! Test initialization
  call inittest()

  ! Read starting values
  open(102, file = pfile, status = "old", action = "read")
  do j = 1, nitem
    read(102, *) (items(j)%par(p), p = 0, items(j)%npar - 1)
  end do
  close(102)

  ! Read starting factor scores
  open(103, file = zfile, status = "old", action = "read")
  do i = 1, nobsn
    read(103, *) (z(i, p), p = 1, nfac)
  end do
  close(103)

  ! Read model constraints. Fixing parameter values overwrites the starting
  ! values read from pfile.
  do
    read(100, '(a)', iostat = rstat) cstr
    if (rstat .gt. 0) then
      stop "input ERROR: Invalid model constraints encountered."
    else if (rstat .lt. 0) then
      exit
    else
      read(cstr, "(a2)") key
      if (key .eq. "fi") then
        read( cstr(3:), * ) j, p, val
        items(j)%fi(p + 1) = p
        items(j)%par(p) = val
      end if
    end if
  end do
  close(100)

  ! item-specific initialization
  do j = 1, nitem
    call inititem(j)
  end do

end subroutine input

! shipout: write parameter estimates (and factor scores) to files
!
! Arguments: none
!
! Effect: 
!   Parameter estimates: "[fstem]-fid.txt"
!   Factor scores: "[fstem]-sco.txt"

subroutine shipout()
  integer :: i, j, p

  open(201, file = trim(fstem)//"-fid.txt", access = "append", action = "write")
  do j = 1, nitem
    do p = 0, items(j)%npar - 1
      write (201, "(f15.8)", advance = "no") items(j)%par(p)
    end do
  end do
  close(201)

  ! order: every line is in natural fortran matrix order
  if (fscore) then
    open(202, file = trim(fstem)//"-sco.txt", access = "append", action = "write")
    do p = 1, nfac
      do i = 1, nobsn
        write (202, "(f15.8)", advance = "no") z(i, p)
      end do
    end do
    close(202)
  end if
end subroutine shipout

end module io
