! test.f90: test configuration
!
! Author: Yang Liu
!
! Last modified: 10/07/2014
!
! Dependency:
!   io: input/output
!   test: test module
!   rng: random number generation module
!   mt19937: a Mersenne Twister RNG

module mcmc
use io
use test
use rng
use mt19937
implicit none
private

! private variables used in rmc
integer, allocatable :: indx(:), indxr(:)  ! 1,...,2 * nobsn
logical, allocatable :: inact(:)  ! observation is active or not

public :: run

contains

! run: The main loop of the MCMC algorithm
!
! Arguments: none
!
! Effect: 
!   Generate a Markov Chain which converges to the joint distribution of A's and
!   Z's conditional on the Q set being non-empty

subroutine run()
  integer :: i, j, r
  !double precision :: t1, t2, t3, t4, ti1, ti2, ti3
  !type(vnode), pointer :: iv
  !type(fnode), pointer :: ia
  !type(enode), pointer :: ie, ie1

  call sgrnd(seed)  ! initialize Mersenne Twister 
  call start()  ! starting values

  ! indx and active face vector
  allocate( inact(2 * nobsn) )
  allocate( indx(2 * nobsn) )
  forall ( i = 1:(2 * nobsn) )
    indx(i) = i
  end forall
  indxr = pack( transpose( reshape(indx, [nobsn, 2]) ), .true. )

  ! main loop
  do r = 1, max_iter
    write(*, "(A, I0, A)", advance="no"), "MCMC cycle ", r, '\r'
    !ti1 = 0.d0; ti2 = 0.d0; ti3 = 0.d0
    do i = 1, nobsn
      !$omp parallel num_threads(4) if(nitem * nfac .gt. 30)
      !$omp do
      !call cpu_time(t1)
      do j = 1, nitem
        call rmc(i, j)
      end do
      !call cpu_time(t2); ti1 = ti1 + t2 - t1
      !$omp end do
      !$omp end parallel
      call gibbsz(i)
      !call cpu_time(t3); ti2 = ti2 + t3 - t2
      !$omp parallel num_threads(4) if(nitem * nfac .gt. 30)
      !$omp do
      do j = 1, nitem
        call getrhs(i, j)
        call getrhs(i + nobsn, j)
        call gibbsa(i, j)
        call update(i, j)
        call getrhs(i + nobsn, j)
        call update(i + nobsn, j)
      end do
      !call cpu_time(t4); ti3 = ti3 + t4 - t3
      !$omp end do
      !$omp end parallel
    end do
    !print *, ti1, ti2, ti3
    ! after burn-in period, and every _thin_ cycles, output parameter estimates
    if ( r .gt. burnin .and. mod( (r - burnin), thin) .eq. 0) then
      call estimate()
      call shipout()
    end if
  end do

  !iv => items(3)%vlist
  !do while ( associated(iv) )
  !  print *, iv%par, iv%obsn
  !  iv => iv%next
  !end do
  !print *, '\n'

  !do i = 24, 24
  !  call rmc(i, 3)
  !end do

  !iv => items(3)%vlist
  !do while ( associated(iv) )
  !  print *, iv%par, iv%obsn
  !  iv => iv%next
  !end do
end subroutine run

! estimate: Apply selection rule.
!
! Argument: none
!
! Effect: For each item, select one vertex from the current interior polytope

subroutine estimate()
  integer :: i, j, which, nv
  type(vnode), pointer :: iv

  do j = 1, nitem
    if ( items(j)%nfr .eq. 0 ) cycle

    ! count number of vertices
    nv = 0
    iv => items(j)%vlist
    do while ( associated(iv) )
      nv = nv + 1
      iv => iv%next
    end do

    ! randomly select a vertex
    which = int( anint( grnd() * ( dble(nv) - 1.d0 ) ) )

    ! take this vertex as the estimate
    iv => items(j)%vlist
    do i = 1, which
      iv => iv%next
    end do
    items(j)%par( items(j)%fr ) = iv%par
  end do
end subroutine estimate

! rmc: Remove a certain constraint from the interior polytope
!
! Argument:
!   obsn: observation index (.le. nobsn)
!   item: item index
!
! Effect:
!   Check if interior polytope is linked with obsn or not; if so, re-initialize
!   vlist and elist, then loop over all observations but the current one to
!   update vertices

subroutine rmc(obsn, item)
  integer, intent(in) :: obsn, item
  type(vnode), pointer :: iv!, fv
  integer :: i
  integer, allocatable :: intf(:), indx1(:), indx2(:)
  !integer :: indx(2 * nobsn)!, rnk(2 * nobsn)
  !double precision :: div, numer, denom, &
  !  d(2 * nobsn), zobsn(0:items(item)%npar - 1)

  ! check if the current vlist is linked with obsn
  iv => items(item)%vlist
  do while ( associated(iv) )
    if( count(iv%obsn .eq. obsn) .gt. 0 .or. &
        count(iv%obsn .eq. obsn + nobsn) .gt. 0 ) then
      !fv => iv
      exit
    end if
    iv => iv%next
  end do

  ! if not linked, end the subroutine
  if ( .not. associated(iv) ) return

  !! method 1: compute distance and then rank
  !! loop over all half-spaces and compute distance to the polytope
  !d = inf
  !do i = 1, 2 * nobsn
  !  ! for obsn to be removed
  !  if (i .eq. obsn .or. i .eq. obsn + nobsn) then
  !    d(i) = -inf
  !    cycle
  !  end if
  !  ! others
  !  zobsn = rtrvz(i, item)
  !  numer = dot_product( items(item)%par( items(item)%fi ), &
  !    zobsn(items(item)%fi) ) - rtrva(i, item)
  !  !denom = 1.d0 / sqrt( sum(zobsn(items(item)%fr)**2) )
  !  !iv => items(item)%vlist
  !  !do while ( associated(iv) )
  !  !  if (iv%lrmf) then
  !  !    div = abs( dot_product( iv%par, zobsn(items(item)%fr) ) &
  !  !      + numer ) * denom
  !  !    if ( div .lt. d(i) ) d(i) = div
  !  !  end if
  !  !  iv => iv%next
  !  !end do
  !  d(i) = abs( dot_product( fv%par, zobsn(items(item)%fr) ) &
  !    + numer ) / sqrt( sum(zobsn(items(item)%fr)**2) )
  !end do

  !! sort observations by distance
  !call mrgrnk(d, rnk)
  !
  !! rebuild interior polytope
  !call reinitv(item)
  !call reinite(item)
  !do i = 3, 2 * nobsn
  !  call getrhs(rnk(i), item)
  !  call update(rnk(i), item)
  !end do

  ! method 2: add interior faces first, then rest
  ! collect all observations linked to the vlist
  allocate( intf(0) )
  iv => items(item)%vlist
  do while ( associated(iv) )
    call union(intf, iv%obsn)
    iv => iv%next
  end do
  inact = .true.
  inact(intf) = .false.
  indx2 = pack( indxr, pack( transpose( reshape( inact, [nobsn, 2]) ), .true. ) )
  inact(obsn) = .true.
  inact(obsn + nobsn) = .true.
  inact = .not. inact
  indx1 = pack(indx, inact)
  call reinitv(item)
  call reinite(item)
  do i = 1, size(indx1)
    call getrhs(indx1(i), item)
    call update(indx1(i), item)
  end do
  do i = 1, size(indx2)
    call getrhs(indx2(i), item)
    call update(indx2(i), item)
  end do
  deallocate(intf)

  !! method 3: sort observations by inactive time
  !! sort observations
  !call mrgrnk(items(item)%ti, rnk)
  !! rebuild interior polytope
  !call reinitv(item)
  !call reinite(item)
  !do i = 1, 2 * nobsn
  !  if (rnk(i) .eq. obsn .or. rnk(i) .eq. obsn + nobsn) cycle
  !  call getrhs(rnk(i), item)
  !  call update(rnk(i), item)
  !end do
end subroutine rmc

! start: Obtain starting values for the underlying response variable A's. It
! requires initial values for the parameter estimates (in pfile), and the latent
! variable Z's (in zfile). It runs the Gibbs sampling step for A.
!
! Arguments: none
!
! Effect:
!   Assign initial values for a(:, :)

subroutine start()
  integer :: i, j
  !type(vnode), pointer :: iv
  !type(enode), pointer :: ie

  do j = 1, nitem
    do i = 1, nobsn
      call getrhs(i, j)
      call getrhs(i + nobsn, j)
      call gibbsa(i, j)
    end do
    call reinitv(j)
    call reinite(j)
    do i = 1, nobsn
      call getrhs(i, j)
      call update(i, j)
      call getrhs(i + nobsn, j)
      call update(i + nobsn, j)
    end do
  end do

  !do i = 1, 1000
  !  !print *, "i =", i
  !  call getrhs(i, 1)
  !  call update(i, 1)
  !  call getrhs(i + nobsn, 1)
  !  call update(i + nobsn, 1)
  !  !iv => items(1)%vlist
  !  !do while ( associated(iv) )
  !  !  print *, iv%par, iv%obsn
  !  !  iv => iv%next
  !  !end do

  !  !ie => items(1)%elist
  !  !do while ( associated(ie) )
  !  !  print *, ie%v1%obsn, ie%v2%obsn, ie%obsn, associated(ie%prev), associated(ie%next)
  !  !  ie => ie%next
  !  !end do
  !end do
end subroutine start

! gibbsz: Gibbs sampling step for latent variable Z's. It generates
! Z from a suitably truncated standard normal distribution.
!
! Arguments:
!   obsn: observation index
!
! Effect:
!   Update z(obsn, :)

subroutine gibbsz(obsn)
  integer, intent(in) :: obsn
  integer :: s, i, j, d, e, m, n, slpd
  double precision :: focal, b, bmin = 0.d0, bmax = 0.d0, aobsn, newz
  double precision, allocatable :: t(:, :), tj(:, :), p(:), zobsn(:)
  type(vnode), pointer :: iv

  fac_loop: do d = 1, nfac
    ! initialize truncation interval: (-inf, inf)
    allocate( t(1, 2) )
    t(1, 1) = - inf; t(1, 2) = inf
    ! loop over items
    item_loop: do j = 1, nitem
      allocate( zobsn(0:items(j)%npar - 1) )
      slpd = items(j)%ncat + d
      lr_loop: do s = 0, nobsn, nobsn
        bmin = inf; bmax = - inf
        aobsn = rtrva(obsn + s, j)
        zobsn = rtrvz(obsn + s, j)
        ! case 1: no vertex (all slopes are fixed)
        if ( .not. associated(items(j)%vlist) ) then
          b = aobsn
          do i = 1, items(j)%nfi
            if ( items(j)%fi(i) .eq. d ) cycle
            b = b - items(j)%par( items(j)%fi(i) ) * zobsn( items(j)%fi(i) )
          end do
          b = b / items(j)%par(slpd)
          if ( (s .eq. 0) .neqv. (items(j)%par(slpd) .lt. 0.d0) ) then
            bmin = b
          else 
            bmax = b
          end if
        ! case 2: if there are vertices
        else
          iv => items(j)%vlist
          ! check if focal slope is fixed
          e = 0
          do i = 1, items(j)%nfr
            if ( items(j)%fr(i) .eq. slpd) then
              e = i
              exit
            end if
          end do
          ! case 2.1: if focal slope is fixed
          if (e .eq. 0) then
            ! if fixed at 0, cycle lr loop
            if ( items(j)%par(slpd) .eq. 0.d0 ) cycle lr_loop
            ! otherwise
            do while ( associated(iv) )
              b = aobsn - dot_product( iv%par, zobsn(items(j)%fr) )
              do i = 1, items(j)%nfi
                if ( items(j)%fi(i) .ne. d ) &
                  b = b - items(j)%par( items(j)%fi(i) ) * zobsn( items(j)%fi(i) )
              end do
              b = b / items(j)%par(slpd)
              ! update bmin and bmax
              if ( (s .eq. 0) .neqv. ( items(j)%par(slpd) .lt. 0.d0 ) ) then
                if (b .lt. bmin) bmin = b
              else 
                if (b .gt. bmax) bmax = b
              end if
              iv => iv%next
            end do
          ! case 2.2: if focal slope is free
          else
            do while ( associated(iv) )
              focal = iv%par(e)
              if (focal .eq. 0.d0) then
                iv => iv%next
                cycle  ! prevent division by 0
              end if
              b = aobsn - dot_product( items(j)%par(items(j)%fi), zobsn(items(j)%fi) )
              do i = 1, items(j)%nfr
                if ( i .ne. e ) &
                  b = b - iv%par(i) * zobsn( items(j)%fr(i) )
              end do
              b = b / focal
              ! update bmin and bmax
              if ( (s .eq. 0) .neqv. (focal .lt. 0.d0) ) then
                if (b .lt. bmin) bmin = b
              else 
                if (b .gt. bmax) bmax = b
              end if
              iv => iv%next
            end do
          end if
        end if
        ! discuss cases for tj
        ! if no bmin
        if ( bmin .ge. inf ) then
          allocate( tj(1, 2) )
          tj(1, 1) = - inf; tj(1, 2) = bmax
        ! if no bmax
        else if ( bmax .le. -inf ) then
          allocate( tj(1, 2) )
          tj(1, 1) = bmin; tj(1, 2) = inf
        else
          ! if bmin <= bmax
          if (bmin .le. bmax) then
            allocate( tj(1, 2) )
            tj(1, 1) = - inf; tj(1, 2) = inf
          ! if bmin > bmax
          else
            allocate( tj(2, 2) )
            tj(1, 1) = - inf; tj(1, 2) = bmax
            tj(2, 1) = bmin; tj(2, 2) = inf
          end if
        end if
        ! update t
        call intersect(t, tj)
        deallocate(tj)
      end do lr_loop
      deallocate(zobsn)
    end do item_loop
    ! generate z
    n = size(t, 1)
    if (n .eq. 1) then
      newz = rtnorm( t(1, 1), t(1, 2), 5 )
      if (newz .lt. inf) z(obsn, d) = newz
    else
      allocate( p(n) )
      do i = 1, n
        p(i) = pnorm( t(i, 2) ) - pnorm( t(i, 1) )
      end do
      m = rmultinom(p)
      newz = rtnorm( t(m, 1), t(m, 2), 5 )
      if (newz .lt. inf) z(obsn, d) = newz
      deallocate(p)
    end if
    deallocate(t)
  end do fac_loop

end subroutine gibbsz

! gibbsa: Gibbs sampling step for underlying response variable A's. It generates
! A from a suitably truncated standard normal distribution.
!
! Arguments:
!   obsn: observation index
!   item: item index
!
! Effect:
!   Update a(obsn, item)

subroutine gibbsa(obsn, item)
  integer, intent(in) :: obsn, item
  a(obsn, item) = rtlogis( items(item)%min_rhs(2), items(item)%max_rhs(1), 5 )
end subroutine gibbsa

! update: Given the current interior vertex and edge lists and a "new"
! observation, update both lists.
! update_1: Update the vertex and edge liests after removing one face.
!
! Arguments:
!   obsn: observation index
!   item: item index
!
! Effect:
!   Remove old infeasible vertices, add new vertices, update the adjacency
!   relationship between vertices

subroutine update(obsn, item)
  integer, intent(in) :: obsn, item
  type(vnode), pointer :: iv, newv!, vv
  type(enode), pointer :: ie, newe, eu
  !type(fnode), pointer :: ia, au
  integer :: obs( items(item)%nfr ), nnew, lr
  !integer, allocatable :: af(:)
  double precision :: aobsn
  !integer :: ct

  ! if no free parameter
  if ( items(item)%nfr .eq. 0 ) return

  ! left or right intercept
  if (obsn .le. 0) then
    stop "update ERROR: cannot update artificial constraints"
  else if (obsn .le. nobsn) then
    lr = 1
  else
    lr = 2
  end if

  ! if the new observation does not cut the current polytope
  aobsn = rtrva(obsn, item)
  if ( ( aobsn .le. items(item)%min_rhs(lr) + 1d-8 ) .or. &
    ( aobsn .ge. items(item)%max_rhs(lr) - 1d-8 ) ) then 
    return
  end if

  ! excurse vlist
  !print *, "markout:", obsn, item
  !print *, rtrva(obsn, item), rtrvz(obsn, item)
  iv => items(item)%vlist
  do while ( associated(iv) )
    ! if the vertex is not feasible any more, mark it
    ! modified to acommodate singular cases (08/01/2014)
    if (aobsn .gt. iv%rhs(lr) - 1d-8 .and. aobsn .lt. iv%rhs(lr) + 1d-8) then
      iv%feas = .false.
    else if (lr .eq. 1 .and. aobsn .le. iv%rhs(lr) - 1d-8) then
      iv%feas = .true.
    else if (lr .eq. 2 .and. aobsn .ge. iv%rhs(lr) + 1d-8) then
      iv%feas = .true.
    else
      iv%feas = .false.
    end if
    iv => iv%next
  end do

  !! allocate new face node
  !allocate(au)
  !au%obsn = obsn
  !au%occr = 0
  !call copy_ve(items(item)%vlist, items(item)%elist, au%v_f, au%e_f)

  ! excurse elist, compute new vertices
  ie => items(item)%elist
  nullify(iv, newv, newe)
  !allocate( af(0) )
  nnew = 0
  ! the while condition here is just for protection when nfr = 1
  do while ( associated(ie) )
    ! if the edge being cut
    obs(1) = obsn; obs( 2:items(item)%nfr ) = ie%obsn
    if ( ie%v1%feas .neqv. ie%v2%feas ) then
      !! occurrences of the faces forming the new vertex += 1 (YL 10/05/2014)
      !! current
      !au%occr = au%occr + 1
      !! others
      !ct = 0
      !ia => items(item)%flist
      !if ( .not. all(ie%obsn .le. 0) ) then
      !  do while ( associated(ia) )
      !    if ( any(ie%obsn .eq. ia%obsn) ) then
      !      ia%obsn = ia%obsn + 1
      !      ct = ct + 1
      !    end if
      !    if (ct .ge. items(item)%nfr - 1) exit
      !    ia => ia%next
      !  end do
      !end if
      ! count active faces
      !call union(af, obs)
      ! compute new vertex
      nnew = nnew + 1
      allocate(iv)
      iv%init = .false.
      allocate( iv%par(items(item)%nfr) )
      call solv(obs, item, iv)
      call pushfront_v(newv, iv)
      ! update edge
      allocate(eu)
      eu = ie
      eu%init = .false.
      if ( .not. ie%v1%feas ) then
        eu%v1 => iv
      else
        eu%v2 => iv
      end if
      call pushfront_e(newe, eu)
      call rmenode( ie, items(item)%elist )
      nullify(iv, eu)
    else if ( .not. ie%v1%feas .and. .not. ie%v2%feas ) then
      call rmenode( ie, items(item)%elist )
    else
      if ( associated(ie%next) ) then
        ie => ie%next
      else 
        exit
      end if
    end if
  end do

  ! Remove infeasible vertices
  iv => items(item)%vlist
  do
    if ( .not. iv%feas ) then
      call rmvnode(iv, items(item)%vlist)
      !! occurrence of faces -= 1
      !ct = 0
      !ia => items(item)%flist
      !do while ( associated(ia) )
      !  if ( any(iv%obsn .eq. ia%obsn) ) then
      !    ia%obsn = ia%obsn - 1
      !    ct = ct + 1
      !  end if
      !  if (ct .ge. items(item)%nfr) exit
      !  ia => ia%next
      !end do
    else
      !call union(af, iv%obsn)
      if ( associated(iv%next) ) then
        iv => iv%next
      else
        exit
      end if
    end if
  end do

  !! update face list
  !ia => items(item)%flist
  !do while ( associated(ia) )
  !  ! remove
  !  if (ia%occr .eq. 0) then
  !    call rmfnode( ia, items(item)%flist )
  !  else
  !    if ( associated(ia%next) ) then
  !      ia => ia%next
  !    else 
  !      exit
  !    end if
  !  end if
  !end do
  !vv => items(item)%vlist
  !do while ( associated(vv) )
  !  print *, vv%obsn
  !  vv => vv%next
  !end do
  !print *,
  !vv => au%v_f 
  !do while ( associated(vv) )
  !  print *, vv%obsn
  !  vv => vv%next
  !end do
  !print *,
  ! add the current face
  !call pushfront_f(items(item)%flist, au)
  !nullify(au)

  ! Append new vertices and new edges
  call edges(item, newv, newe, nnew)
  !print *, "olde:"
  !eu => items(item)%elist
  !do while ( associated(eu) )
  !  print *, eu%v1%obsn, eu%v2%obsn, eu%obsn
  !  eu => eu%next
  !end do
  !print *, "newe:"
  !eu => newe
  !do while ( associated(eu) )
  !  print *, eu%v1%obsn, eu%v2%obsn, eu%obsn
  !  eu => eu%next
  !end do
  iv%next => newv
  newv%prev => iv
  ! when nfr = 1, ie is null
  if ( .not. associated(ie) ) then
    items(item)%elist => newe
  else
    ie%next => newe
    newe%prev => ie
  end if

  ! count active faces
  !af = pack( af, af .gt. 0 )
  !items(item)%act = .false.
  !items(item)%act(af) = .true.
  !where ( items(item)%act )
  !  items(item)%ti = 0
  !end where
  !where ( .not. items(item)%act )
  !  items(item)%ti = items(item)%ti + 1
  !end where

end subroutine update

!subroutine update_1(obsn, item, vlist, elist, mirhs, marhs)
!  integer, intent(in) :: obsn, item
!  double precision, intent(in) :: mirhs(2), marhs(2)
!  type(vnode), intent(inout), pointer :: vlist
!  type(enode), intent(inout), pointer :: elist
!  type(vnode), pointer :: iv, newv
!  type(enode), pointer :: ie, newe, eu
!  integer :: obs( items(item)%nfr ), nnew, lr
!  double precision :: aobsn
!
!  ! if no free parameter
!  if ( items(item)%nfr .eq. 0 ) return
!
!  ! left or right intercept
!  if (obsn .le. 0) then
!    stop "update ERROR: cannot update artificial constraints"
!  else if (obsn .le. nobsn) then
!    lr = 1
!  else
!    lr = 2
!  end if
!
!  ! if the new observation does not cut the current polytope
!  aobsn = rtrva(obsn, item)
!  if ( ( aobsn .le. mirhs(lr) + 1d-8 ) .or. &
!    ( aobsn .ge. marhs(lr) - 1d-8 ) ) then 
!    return
!  end if
!
!  ! excurse vlist, mark feasibility
!  iv => vlist
!  do while ( associated(iv) )
!    ! if the vertex is not feasible any more, mark it
!    ! modified to acommodate singular cases (08/01/2014)
!    if (aobsn .gt. iv%rhs(lr) - 1d-8 .and. aobsn .lt. iv%rhs(lr) + 1d-8) then
!      iv%feas = .false.
!    else if (lr .eq. 1 .and. aobsn .le. iv%rhs(lr) - 1d-8) then
!      iv%feas = .true.
!    else if (lr .eq. 2 .and. aobsn .ge. iv%rhs(lr) + 1d-8) then
!      iv%feas = .true.
!    else
!      iv%feas = .false.
!    end if
!    iv => iv%next
!  end do
!
!  ! excurse elist, compute new vertices
!  ie => elist
!  nullify(iv, newv, newe)
!  nnew = 0
!  ! the while condition here is just for protection when nfr = 1
!  do while ( associated(ie) )
!    ! if the edge being cut
!    obs(1) = obsn; obs( 2:items(item)%nfr ) = ie%obsn
!    if ( ie%v1%feas .neqv. ie%v2%feas ) then
!      ! compute new vertex
!      nnew = nnew + 1
!      allocate(iv)
!      iv%init = .false.
!      allocate( iv%par(items(item)%nfr) )
!      call solv(obs, item, iv)
!      call pushfront_v(newv, iv)
!      ! update edge
!      allocate(eu)
!      eu = ie
!      eu%init = .false.
!      if ( .not. ie%v1%feas ) then
!        eu%v1 => iv
!      else
!        eu%v2 => iv
!      end if
!      call pushfront_e(newe, eu)
!      call rmenode( ie, elist )
!      nullify(iv, eu)
!    else if ( .not. ie%v1%feas .and. .not. ie%v2%feas ) then
!      call rmenode( ie, elist )
!    else
!      if ( associated(ie%next) ) then
!        ie => ie%next
!      else 
!        exit
!      end if
!    end if
!  end do
!
!  ! Remove infeasible vertices
!  iv => vlist
!  do
!    if ( .not. iv%feas ) then
!      call rmvnode(iv, vlist)
!    else
!      if ( associated(iv%next) ) then
!        iv => iv%next
!      else
!        exit
!      end if
!    end if
!  end do
!
!  ! Append new vertices and new edges
!  call edges(item, newv, newe, nnew)
!  iv%next => newv
!  newv%prev => iv
!  ! when nfr = 1, ie is null
!  if ( .not. associated(ie) ) then
!    elist => newe
!  else
!    ie%next => newe
!    newe%prev => ie
!  end if
!
!end subroutine update_1

! getrhs: Assign the rhs values in a vnode linked-list, for fixed observation
! i and item j. The min and max rhs values are also returned.
!
! Arguments:
!   obsn: observation index
!   item: item index
!
! Effect:
!   Assign rhs in each vnode, and min_rhs, max_rhs in items(item)

subroutine getrhs(obsn, item)
  integer, intent(in) :: obsn, item
  type(vnode), pointer :: iv
  integer :: lr
  double precision :: zobsn(0:items(item)%npar - 1)

  ! left or right intercept
  if (obsn .le. 0) then
    stop "getrhs ERROR: cannot compute rhs for artificial constraints"
  else if (obsn .le. nobsn) then
    lr = 1
  else
    lr = 2
  end if

  zobsn = rtrvz(obsn, item)
  ! when all parameters are fixed
  if ( .not. associated(items(item)%vlist) ) then
    items(item)%max_rhs(lr) = dot_product( items(item)%par( items(item)%fi ), &
      zobsn(items(item)%fi) )
    items(item)%min_rhs(lr) = items(item)%max_rhs(lr)
    return
  end if
  ! otherwise
  iv => items(item)%vlist
  items(item)%min_rhs(lr) = inf
  items(item)%max_rhs(lr) = - inf
  do while ( associated(iv) )
    iv%rhs(lr) = dot_product( items(item)%par( items(item)%fi ), &
      zobsn(items(item)%fi) ) + &
      dot_product( iv%par, zobsn(items(item)%fr) )
    if ( iv%rhs(lr) .gt. items(item)%max_rhs(lr) ) &
      items(item)%max_rhs(lr) = iv%rhs(lr)
    if ( iv%rhs(lr) .lt. items(item)%min_rhs(lr) ) &
      items(item)%min_rhs(lr) = iv%rhs(lr)
    iv => iv%next
  end do

end subroutine getrhs

end module mcmc
