! rng.f90: Functions and subroutines for random number generation. The
! fundamantal uniform(0, 1) generator is included in mt19937.f90.
!
! Author: Yang Liu
!
! Last modified: 08/28/2014
!
! Contains:
!   rng: Random data generation
!   mt19937: A Mersenne Twister RNG

! MT19937, Mersenne Twister Random Number Generator ([0,1) Real Number)
! in Fortran90
!
! Usage: 
!   1) When you use mt19937, add the sentence "use mt19937" 
!      above the implicit sentence.
!   2) To set an initial seed, call sgrnd(seed). (The "seed" is an integer.)
!      If you do not call this, the seed is 4357.
!   3) Use the function grnd().
!      (Do not declare "real(8) :: grnd".)
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! A C-program for MT19937: Real number version
!   genrand() generates one pseudorandom real number (double)
! which is uniformly distributed on [0,1]-interval, for each
! call. sgenrand(seed) set initial values to the working area
! of 624 words. Before genrand(), sgenrand(seed) must be
! called once. (seed is any 32-bit integer except for 0).
! Integer generator is obtained by modifying two lines.
!   Coded by Takuji Nishimura, considering the suggestions by
! Topher Cooper and Marc Rieffel in July-Aug. 1997.
!
! This library is free software; you can redistribute it and/or
! modify it under the terms of the GNU Library General Public
! License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later
! version.
! This library is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
! See the GNU Library General Public License for more details.
! You should have received a copy of the GNU Library General
! Public License along with this library; if not, write to the
! Free Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
! 02111-1307  USA
!
! Copyright (C) 1997 Makoto Matsumoto and Takuji Nishimura.
! When you use this, send an email to: matumoto@math.keio.ac.jp
! with an appropriate reference to your work.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Fortran translation by Hiroshi Takano.  Jan. 13, 1999.
!
!   genrand()      -> double precision function grnd()
!   sgenrand(seed) -> subroutine sgrnd(seed)
!                     integer seed
!
! This program uses the following non-standard intrinsics.
!   ishft(i,n): If n>0, shifts bits in i by n positions to left.
!               If n<0, shifts bits in i by n positions to right.
!   iand (i,j): Performs logical AND on corresponding bits of i and j.
!   ior  (i,j): Performs inclusive OR on corresponding bits of i and j.
!   ieor (i,j): Performs exclusive OR on corresponding bits of i and j.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Fortran90 translation by Yusuke Konishi. May. 13, 2013.
!
! The interval is changed from [0, 1] to [0, 1).
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
module mt19937
  implicit none
  integer, private :: N, N1, M, MATA, UMASK, LMASK, TMASKB, TMASKC
  parameter( &
             & N = 624, &
             & N1 = 625, &
             & M = 397, &
             & MATA = -1727483681, &
             & UMASK = -2147483648, &
             & LMASK = 2147483647, &
             & TMASKB = -1658038656, &
             & TMASKC = -272236544 &
             & )
  integer,private :: mti = N1, mt(0:N-1), mag01(0:1) = (/0, MATA/)

  contains

  subroutine sgrnd(seed)
    integer,intent(in) :: seed
!
!      setting initial seeds to mt[N] using
!      the generator Line 25 of Table 1 in
!      [KNUTH 1981, The Art of Computer Programming
!         Vol. 2 (2nd Ed.), pp102]
!
    mt(0) = iand(seed, -1)
    do mti = 1, N - 1
      mt(mti) = iand(69069 * mt(mti - 1), -1)
    end do
  end subroutine sgrnd

  real(8) function grnd()
    integer :: y, kk

    if(mti >= N) then
!                   generate N words at one time
      if(mti == N + 1) then
!                        if sgrnd() has not been called,
        call sgrnd(4357)
!                          a default initial seed is used
      endif

      do kk = 0, N - M - 1
        y = ior(iand(mt(kk), UMASK), iand(mt(kk + 1), LMASK))
        mt(kk) = ieor(ieor(mt(kk + M), ishft(y, -1)), mag01(iand(y, 1)))
      end do

      do kk = N - M, N - 2
        y = ior(iand(mt(kk), UMASK), iand(mt(kk + 1), LMASK))
        mt(kk) = ieor(ieor(mt(kk + (M - N)), ishft(y, -1)), mag01(iand(y, 1)))
      end do

      y = ior(iand(mt(N - 1), UMASK), iand(mt(0), LMASK))
      mt(N - 1) = ieor(ieor(mt(M - 1), ishft(y, -1)), mag01(iand(y, 1)))
      mti = 0
    endif

    y = mt(mti)
    mti = mti + 1
    y = ieor(y, ishft(y, -11))
    y = ieor(y, iand(ishft(y, 7), TMASKB))
    y = ieor(y, iand(ishft(y, 15), TMASKC))
    y = ieor(y, ishft(y, -18))

    if(y < 0) then
      grnd = (dble(y) + 2.0d0 ** 32) / (2.0d0 ** 32)
    else
      grnd = dble(y) / (2.0d0 ** 32)
    endif
  end function grnd

end module mt19937
!   mt19937: A Mersenne Twister random number generator

module rng
use mt19937
implicit none
public

! private variables
double precision, parameter, private :: pi = 3.141592653589793D+00

contains

! pnorm: Standard normal cdf function. Originally Hart's algorithm 5666, with
! the first 15 digits being precise.
!
! Author: 
!   Alan Miller (fortran 77)
!   John Burkardt (fortran 90)
!
! Last modified: 01/13/2008
!
! Licensing: This code is distributed under the GNU LGPL license.
!
!  Reference:
!    John Hart, Ward Cheney, Charles Lawson, Hans Maehly,
!    Charles Mesztenyi, John Rice, Henry Thacher,
!    Christoph Witzgall,
!    Computer Approximations,
!    Wiley, 1968,
!    LC: QA297.C64.
!
!  Arguments:
!    x: x value where the cdf function is evaluated
!
!  Return:
!    p: cdf value
 
function pnorm(z) result(p)
  implicit none

  double precision, intent(in) :: z
  double precision :: p
  double precision, parameter :: cutoff = 7.071D+00
  double precision, parameter :: p0 = 220.2068679123761D+00
  double precision, parameter :: p1 = 221.2135961699311D+00
  double precision, parameter :: p2 = 112.0792914978709D+00
  double precision, parameter :: p3 = 33.91286607838300D+00
  double precision, parameter :: p4 = 6.373962203531650D+00
  double precision, parameter :: p5 = 0.7003830644436881D+00
  double precision, parameter :: p6 = 0.03526249659989109D+00
  double precision, parameter :: q0 = 440.4137358247522D+00
  double precision, parameter :: q1 = 793.8265125199484D+00
  double precision, parameter :: q2 = 637.3336333788311D+00
  double precision, parameter :: q3 = 296.5642487796737D+00
  double precision, parameter :: q4 = 86.78073220294608D+00
  double precision, parameter :: q5 = 16.06417757920695D+00
  double precision, parameter :: q6 = 1.755667163182642D+00
  double precision, parameter :: q7 = 0.08838834764831844D+00
  double precision, parameter :: root2pi = 2.506628274631001D+00
  double precision :: expntl, zabs, pdf

  zabs = abs(z)
  ! 37 < |Z|.
  if ( 37.0D+00 .lt. zabs ) then
    p = 0.0D+00
  ! |Z| <= 37.
  else
    expntl = exp ( - 0.5D+00 * zabs * zabs )
    pdf = expntl / root2pi
    ! |Z| < CUTOFF = 10 / sqrt(2).
    if ( zabs .lt. cutoff ) then
      p = expntl * (((((( &
          p6   * zabs &
        + p5 ) * zabs &
        + p4 ) * zabs &
        + p3 ) * zabs &
        + p2 ) * zabs &
        + p1 ) * zabs &
        + p0 ) / ((((((( &
          q7   * zabs &
        + q6 ) * zabs &
        + q5 ) * zabs &
        + q4 ) * zabs &
        + q3 ) * zabs &
        + q2 ) * zabs &
        + q1 ) * zabs &
      + q0 )
    ! CUTOFF <= |Z|.
    else
      p = pdf / ( &
        zabs + 1.0D+00 / ( &
        zabs + 2.0D+00 / ( &
        zabs + 3.0D+00 / ( &
        zabs + 4.0D+00 / ( &
        zabs + 0.65D+00 )))))
    end if
  end if

  if ( z .gt. 0.0D+00 ) p = 1.d0 - p
end function pnorm

! qnorm: Invert the standard normal CDF. The result is accurate to about 1 part
! in 10**16.
!
! Licensing: This code is distributed under the GNU LGPL license.
!
! Modified: 06/05/2007
!
! Author: 
!   Michael Wichura (fortran 77 version)
!   John Burkardt (fortran 90 version)
!
! Reference:
!   Michael Wichura,
!   Algorithm AS241:
!   The Percentage Points of the Normal Distribution,
!   Applied Statistics,
!   Volume 37, Number 3, pages 477-484, 1988.
!
! Arguments:
!   p: probability value from [0, 1]
!
! Return:
!   x: standard normal quantile

function qnorm(p) result(x)

  double precision, parameter, dimension ( 8 ) :: a = (/ &
    3.3871328727963666080D+00, &
    1.3314166789178437745D+02, &
    1.9715909503065514427D+03, &
    1.3731693765509461125D+04, &
    4.5921953931549871457D+04, &
    6.7265770927008700853D+04, &
    3.3430575583588128105D+04, &
    2.5090809287301226727D+03 /)
  double precision, parameter, dimension ( 8 ) :: b = (/ &
    1.0D+00, &
    4.2313330701600911252D+01, &
    6.8718700749205790830D+02, &
    5.3941960214247511077D+03, &
    2.1213794301586595867D+04, &
    3.9307895800092710610D+04, &
    2.8729085735721942674D+04, &
    5.2264952788528545610D+03 /)
  double precision, parameter, dimension ( 8 ) :: c = (/ &
    1.42343711074968357734D+00, &
    4.63033784615654529590D+00, &
    5.76949722146069140550D+00, &
    3.64784832476320460504D+00, &
    1.27045825245236838258D+00, &
    2.41780725177450611770D-01, &
    2.27238449892691845833D-02, &
    7.74545014278341407640D-04 /)
  double precision, parameter :: const1 = 0.180625D+00
  double precision, parameter :: const2 = 1.6D+00
  double precision, parameter, dimension ( 8 ) :: d = (/ &
    1.0D+00, &
    2.05319162663775882187D+00, &
    1.67638483018380384940D+00, &
    6.89767334985100004550D-01, &
    1.48103976427480074590D-01, &
    1.51986665636164571966D-02, &
    5.47593808499534494600D-04, &
    1.05075007164441684324D-09 /)
  double precision, parameter, dimension ( 8 ) :: e = (/ &
    6.65790464350110377720D+00, &
    5.46378491116411436990D+00, &
    1.78482653991729133580D+00, &
    2.96560571828504891230D-01, &
    2.65321895265761230930D-02, &
    1.24266094738807843860D-03, &
    2.71155556874348757815D-05, &
    2.01033439929228813265D-07 /)
  double precision, parameter, dimension ( 8 ) :: f = (/ &
    1.0D+00, &
    5.99832206555887937690D-01, &
    1.36929880922735805310D-01, &
    1.48753612908506148525D-02, &
    7.86869131145613259100D-04, &
    1.84631831751005468180D-05, &
    1.42151175831644588870D-07, &
    2.04426310338993978564D-15 /)
  double precision, intent(in) :: p
  double precision :: q
  double precision :: r
  double precision, parameter :: split1 = 0.425D+00
  double precision, parameter :: split2 = 5.0D+00
  double precision :: x

  if ( p <= 0.0D+00 ) then
    x = - huge ( x )
    return
  end if
  if ( 1.0D+00 <= p ) then
    x = huge ( x )
    return
  end if
  q = p - 0.5D+00
  if ( abs ( q ) <= split1 ) then
    r = const1 - q * q
    x = q * poly_value ( 8, a, r ) / poly_value ( 8, b, r )
  else
    if ( q < 0.0D+00 ) then
      r = p
    else
      r = 1.0D+00 - p
    end if
    if ( r <= 0.0D+00 ) then
      x = huge ( x )
    else
      r = sqrt ( - log ( r ) )
      if ( r <= split2 ) then
        r = r - const2
        x = poly_value ( 8, c, r ) / poly_value ( 8, d, r )
      else
        r = r - split2
        x = poly_value ( 8, e, r ) / poly_value ( 8, f, r )
      end if
    end if
    if ( q < 0.0D+00 ) then
      x = -x
    end if
  end if
end function qnorm

! poly_value: Evaluate a polynomial function. For sanity's sake, the value of N
! indicates the NUMBER of coefficients, or more precisely, the ORDER of the
! polynomial, rather than the DEGREE of the polynomial.  The two quantities
! differ by 1, but cause a great deal of confusion.
!
! Given N and A, the form of the polynomial is:
! p(x) = a(1) + a(2) * x + ... + a(n-1) * x^(n-2) + a(n) * x^(n-1)
!
! Licensing: This code is distributed under the GNU LGPL license.
!
! Modified: 08/13/2004
!
! Author: John Burkardt
!
! Arguments:
!   n: order of the polynomial
!   a(n): the coefficients of the polynomial.
!   x: the point at which the polynomial is to be evaluated.
!   poly_value: the value of the polynomial at x.

function poly_value(n, a, x)
  implicit none

  integer :: n
  double precision :: a(n)
  integer :: i
  double precision :: poly_value
  double precision :: x

  poly_value = a(n)
  do i = n - 1, 1, -1
    poly_value = poly_value * x + a(i)
  end do
end function poly_value

! rnorm: samples from the standard normal probability distribution. The
! standard normal probability distribution function (PDF) has mean 0 and
! standard deviation 1. The Box-Muller method is used, which is efficient, but
! generates two values at a time.
!
! Licensing: This code is distributed under the GNU LGPL license.
!
! Modified: 02/02/2002
!
! Author: John Burkardt
!
! Arguments: none
!
! Return: a sample from N(0, 1)

function rnorm() result(x)
  implicit none

  double precision :: x
  double precision, save :: y = 0.0d+00
  logical, save :: used = .true.
  double precision :: r1
  double precision :: r2

  ! If we've used an even number of values so far, generate two more, return one
  ! and save one.
  if (used) then
    do
      r1 = grnd()
      if ( r1 .ne. 0.d0 ) then
        exit
      end if
    end do
    r2 = grnd()
    x = sqrt ( -2.d0 * log ( r1 ) ) * cos ( 2.d0 * pi * r2 )
    y = sqrt ( -2.d0 * log ( r1 ) ) * sin ( 2.d0 * pi * r2 )
    used = .false.
  ! Otherwise, return the second saved value.
  else
    x = y
    used = .true.
  end if
end function rnorm

! rtnorm: Samples from the a standard normal distribution truncated to an
! interval [lb, ub]. Slice sampling is used.
!
! Arguments:
!   lb: lower bound
!   ub: lower bound
!   iter: number of iterations
!
! Return:
!   x: an instance from the truncated normal distribution

function rtnorm(lb, ub, iter) result(x)
  implicit none

  double precision, intent(in) :: lb, ub
  integer, intent(in) :: iter
  double precision :: x, y, h
  double precision :: vu, hl, hu
  integer :: i

  ! if lb >= ub
  if (lb .ge. ub) then
    !print *, "lb =", lb, "ub =", ub
    !stop "rtnorm ERROR: The lower bound is greater than or equal to the upper bound."
    x = huge(1.d0)
    return
  end if
  ! start from the largest vertical slice (just in case someone sets iter = 1)
  if ( (lb .lt. 0.d0) .and. (ub .gt. 0.d0) ) then
    x = 0.d0
  else if (lb .gt. 0.d0) then
    x = lb
  else
    x = ub
  end if
  ! slice sampling
  do i = 1, iter
    vu = exp(-5.d-1 * x * x)
    ! uniform sampling on the vertical slice
    y = grnd() * vu
    h = sqrt( -2.d0 * log(y) ) 
    hl = max(lb, -h)
    hu = min(ub, h)
    ! uniform sampling on the horizontal slice
    x = hl + grnd() * (hu - hl)
  end do

end function rtnorm

! rtlogis: Samples from the a standard logistic distribution truncated to an
! interval [lb, ub]. Slice sampling is used.
!
! Arguments:
!   lb: lower bound
!   ub: lower bound
!   iter: number of iterations
!
! Return:
!   x: an instance from the truncated normal distribution

function rtlogis(lb, ub, iter) result(x)
  double precision, intent(in) :: lb, ub
  integer, intent(in) :: iter
  !double precision :: pl, pu, p
  double precision :: x, y, h, e
  double precision :: vu, hl, hu
  integer :: i

  ! if lb >= ub
  if (lb .ge. ub) then
    print *, "lb =", lb, "ub =", ub
    stop "rtlogis ERROR: The lower bound is greater than or equal to the upper bound."
  end if

  ! inverse cdf method
  !pl = 1.d0 / ( 1.d0 + exp(-lb) )
  !pu = 1.d0 / ( 1.d0 + exp(-ub) )
  !p = pl + grnd() * (pu - pl)
  !x = log( p / (1.d0 - p) )

  ! start from the largest vertical slice (just in case someone sets iter = 1)
  if ( (lb .lt. 0.d0) .and. (ub .gt. 0.d0) ) then
    x = 0.d0
  else if (lb .gt. 0.d0) then
    x = lb
  else
    x = ub
  end if
  ! slice sampling
  do i = 1, iter
    e = exp(x)
    vu = e / (1 + e) ** 2
    ! uniform sampling on the vertical slice
    y = grnd() * vu
    h = log( ( 1.d0 + sqrt(1.d0 - 4.d0 * y) ) / &
      (2.d0 * y) - 1.d0)
    hl = max(lb, -h)
    hu = min(ub, h)
    ! uniform sampling on the horizontal slice
    x = hl + grnd() * (hu - hl)
  end do

end function rtlogis

! rmultinom: Samples from a multinomial distribution
!
! Arguments:
!   p: probability vector
!
! Return:
!   x: an instance from multinomial(p)

function rmultinom(p) result(x)
  double precision, allocatable, intent(in) :: p(:)
  double precision, allocatable :: c(:)
  double precision :: u
  integer :: i, n, x

  n = size(p)
  x = 0
  ! cumulative probability
  allocate( c(0:n), source = 0.d0 )
  do i = 1, n
    if ( p(i) .lt. 0 ) &
      stop "rnultinom ERROR: negative probability found."
    c(i) = c(i - 1) + p(i)
  end do

  ! if c(n) = 0, assign equal probability
  if ( c(n) .eq. 0 ) then
    u = grnd()
    x = 1 + nint( u * (n - 1) )
    return
  ! otherwise, normalize
  else
    c = c / c(n)
  end if
  
  u = grnd()
  do i = 1, n
    if ( ( c(i - 1) .le. u ) .and. ( u .lt. c(i) ) ) then
      x = i
      return
    end if
  end do
end function rmultinom

end module rng
