! test.f90: test configuration
!
! Author: Yang Liu
!
! Last modified: 10/07/2014
!
! Dependency:
!   util: utility functions

module test
use util
implicit none
public

! Global variables:
!   nobsn: number of observation
!   nitem: number of items
!   nfac: number of latent factors
!   seed: seed for RNG
!   max_iter: maximum iterations
!   burnin: number of burn-in cycles
!   thin: thinning interval
!   param_bnd: arbitrary parameter bound
!   y: data
!   z: latent variables
!   a: underlying response variables

integer :: nobsn, nitem
integer :: nfac
integer :: seed, max_iter, burnin, thin
double precision :: param_bnd
integer, allocatable :: y(:, :)
double precision, allocatable :: z(:, :), a(:, :)

! Other public variables
!   nac: number of artificial constraints 2 * (nfac + 1)
!   inf: largest real number
integer :: nac
double precision :: inf

! Defined-type vnode (linked-list node)
type vnode
  type(vnode), pointer :: next, prev
  integer, allocatable :: obsn(:)  ! determining observations
  logical :: init  ! initial vertex?
  logical :: feas  ! feasibility of a vertex
  !logical :: lrmf  ! linked with the face being removed? (YL 10/07/2014)
  double precision, allocatable :: par(:)  ! vertex
  double precision :: rhs(2)  ! right-hand side of the inequality
end type vnode

! Defined-type enode
type enode
  type(enode), pointer :: next, prev
  type(vnode), pointer :: v1, v2
  logical :: init  ! initial edges?
  integer, allocatable :: obsn(:)  ! shared observations
end type enode

!! Defined-type fnode
!type fnode
!  type(fnode), pointer :: next, prev
!  integer :: obsn  ! observation (scalar integer)
!  integer :: occr  ! number of occurrence in vlist
!  type(vnode), pointer :: v_f => null()  ! vertex list if face is removed
!  type(enode), pointer :: e_f => null()  ! edge list if face is removed
!  integer :: min_rhs(2), max_rhs(2)  ! max and min lhs values for v_f
!end type fnode

! Defined-type vpt
type vpt
  type(vnode), pointer :: v
end type vpt

! Defined-type ept
type ept
  type(enode), pointer :: e
end type ept

! Defined-type item
type item
  integer, allocatable :: fi(:), fr(:)  ! fixed and free parameters
  integer, allocatable :: fr_int(:), fr_slp(:)  ! free intercepts and slopes
  integer :: ncat   ! number of response categories
  integer :: npar, nfi, nfr  ! number of all, fixed and free parameters
  integer :: nfr_int, nfr_slp  ! number of free intercepts and slopes
  integer :: nac, nav, nae  ! number of artificial constraints, vertices, edges
  double precision, allocatable :: par(:)  ! parameters
  type(vnode), pointer :: vlist  ! vertex lists
  type(enode), pointer :: elist  ! edge lists
  !type(fnode), pointer :: flist  ! face lists
  type(vpt), allocatable :: vinit(:)  ! initial vnode pointers
  type(ept), allocatable :: einit(:)  ! initial enode pointers
  double precision :: min_rhs(2), max_rhs(2)  ! min and max of rhs values
  !logical, allocatable :: act(:) ! which observation is active
  !integer, allocatable :: ti(:)  ! number of update steps each observation stays inactive
end type item

type(item), allocatable :: items(:)  ! a vector of items 

contains

! inittest: Test initialization
! inititem: Item initialization
! initpoly: Polytope initialization
!
! Arguments: 
!   item: item index
!
! Effect: Assign misc. elements for each item

subroutine inittest()
  inf = huge(1.d0)
  allocate( y(nobsn, nitem) )
  allocate( a(nobsn, nitem) )
  allocate( z(nobsn, nfac) )
  allocate( items(nitem) )
end subroutine inittest

subroutine inititem(item)
  integer, intent(in) :: item
  integer :: p
  ! Initialize elements in type item
  allocate( items(item)%fi( items(item)%npar ), source = -1 )
  allocate( items(item)%fr( items(item)%npar ) )
  do p = 1, items(item)%npar
    items(item)%fr(p) = p - 1
  end do
  allocate( items(item)%par( 0:items(item)%npar - 1 ) )
  !allocate( items(item)%act(nobsn * 2) )
  !allocate( items(item)%ti(nobsn * 2), source = 0 )
end subroutine inititem

subroutine initpoly(item)
  integer, intent(in) :: item
  integer :: p, q, i, r, j
  integer :: nfr_intp1, bnfr_slp
  type(vnode), pointer :: iv
  type(enode), pointer :: ie
  type(ept), allocatable :: einit_tmp(:)
  logical, allocatable :: dup(:)

  ! Finalize counting fixed and free components
  items(item)%fr = pack(items(item)%fr, items(item)%fi .eq. -1)
  items(item)%fi = pack(items(item)%fi, items(item)%fi .ne. -1)
  items(item)%fr_int = pack(items(item)%fr, items(item)%fr .le. items(item)%ncat)
  items(item)%fr_slp = pack(items(item)%fr, items(item)%fr .gt. items(item)%ncat)
  items(item)%nfr = size(items(item)%fr)
  items(item)%nfi = size(items(item)%fi)
  items(item)%nfr_int = size(items(item)%fr_int)
  items(item)%nfr_slp = size(items(item)%fr_slp)
  nfr_intp1 = items(item)%nfr_int + 1
  bnfr_slp = 2 ** items(item)%nfr_slp

  ! Dimensions of artificial constraints, vertices, and edges
  ! artificial constraints order: int_lower_bound, int_order_constraints,
  ! int_upper_bound, slp_lower_upper_bound_duets
  items(item)%nac = nfr_intp1 + 2 * items(item)%nfr_slp
  items(item)%nav = nfr_intp1 * bnfr_slp
  ! for nae, we assign an upper bound first
  items(item)%nae = items(item)%nav * (items(item)%nav - 1) / 2

  ! Prepare initial vertex list
  allocate( items(item)%vinit( items(item)%nav ) )
  j = 0
  do r = 1, nfr_intp1
    do q = 1, bnfr_slp
      j = j + 1
      allocate(iv)
      iv%init = .true.
      allocate( iv%obsn( items(item)%nfr ), &
        iv%par( items(item)%nfr ) )
      i = 0
      ! index for free intercepts
      do p = 1, nfr_intp1
        if (p .eq. r) cycle
        i = i + 1
        iv%obsn(i) = 1 - p
      end do
      ! index for free slopes
      do p = 1, items(item)%nfr_slp
        i = - 2 * p - ibits(q - 1, p - 1, 1)
        iv%obsn(nfr_intp1 + items(item)%nfr_slp - p) = i - &
          items(item)%nfr_int + 1
      end do
      call solv(iv%obsn, item, iv)
      nullify(iv%next)
      if (j .gt. 1) then
        iv%prev => items(item)%vinit(j - 1)%v
        iv%prev%next => iv
      else
        nullify(iv%prev)
      end if
      items(item)%vinit(j)%v => iv
      nullify(iv)
    end do
  end do

  ! vlist contains a single vertex at the begining
  nullify( items(item)%vlist )
  allocate(iv)
  iv%init = .false.
  allocate( iv%obsn( items(item)%nfr ), &
    iv%par( items(item)%nfr ) )
  iv%par = items(item)%par( items(item)%fr )
  call pushfront_v( items(item)%vlist, iv )

  ! Prepare initial edge list
  nullify( items(item)%elist )
  allocate( dup(items(item)%nfr) )
  allocate( items(item)%einit( items(item)%nae ) )
  i = 0
  do p = 2, items(item)%nav
    do q = 1, p - 1
      !dup = .false.
      call duplicate(items(item)%vinit(p)%v%obsn, &
        items(item)%vinit(q)%v%obsn, dup)
      if ( count(dup) .eq. items(item)%nfr - 1 ) then  ! if adjacent
        i = i + 1
        allocate(ie)
        ie%init = .true.
        ie%v1 => items(item)%vinit(p)%v
        ie%v2 => items(item)%vinit(q)%v
        allocate( ie%obsn(items(item)%nfr - 1) )
        ie%obsn = pack(items(item)%vinit(p)%v%obsn, dup)
        nullify(ie%next)
        if (i .gt. 1) then
          ie%prev => items(item)%einit(i - 1)%e
          ie%prev%next => ie
        else
          nullify(ie%prev)
        end if
        items(item)%einit(i)%e => ie
        nullify(ie)
      end if
    end do
  end do
  items(item)%nae = i  ! update nae
  ! remove trailing empty pointers
  einit_tmp = items(item)%einit 
  deallocate(items(item)%einit); allocate( items(item)%einit(i) )
  items(item)%einit = einit_tmp(1:i)
  deallocate(einit_tmp)

  !nullify( items(item)%flist )

end subroutine initpoly

! edges: Get all edges from a linked-list of vnode objects. Computational
! complexity is O(n ** 2) where n is the length of the vertex list
!
! Arguments:
!   item: item index
!   v: a vnode pointer
!   nv: total number of vertices
!
! Output:
!   e: a head enode pointer

subroutine edges(item, v, e, nv)
  integer, intent(in) :: item, nv
  type(vnode), pointer, intent(in) :: v
  type(enode), pointer, intent(inout) :: e
  type(vnode), pointer :: iv1, iv2
  type(enode), pointer :: ie
  logical :: dup( items(item)%nfr )
  integer :: nadj(nv), c1, c2, max_adj
 
  if ( .not. associated(v%next) ) return
  max_adj = items(item)%nfr - 1  ! maximum number of adjacent vertices
  nadj = 0  ! initialize number of adjacencies
  iv2 => v%next
  c2 = 1
  do while ( associated(iv2) )
    iv1 => v
    c2 = c2 + 1
    c1 = 0
    do while ( .not. associated(iv1, iv2) )
      c1 = c1 + 1
      ! if counter nadj(c2) reaches maximum, exit the inner loop
      if ( nadj(c2) .ge. max_adj ) then
        exit
      end if
      ! if counter nadj(c1) reaches maximum, cycle the inner loop
      if ( nadj(c1) .ge. max_adj ) then
        iv1 => iv1%next
        cycle
      end if
      ! otherwise, check adjacency
      !dup = .false.
      call duplicate(iv1%obsn, iv2%obsn, dup)
      if ( count(dup) .eq. items(item)%nfr - 1 ) then  ! if adjacent
        nadj(c1) = nadj(c1) + 1
        nadj(c2) = nadj(c2) + 1
        allocate(ie)
        ie%init = .false.
        ie%v1 => iv1
        ie%v2 => iv2
        allocate( ie%obsn(items(item)%nfr - 1) )
        ie%obsn = pack(iv1%obsn, dup)
        call pushfront_e(e, ie)
        nullify(ie) 
      end if
      iv1 => iv1%next
    end do
    iv2 => iv2%next
  end do
end subroutine edges

! rtrva: retrieve a (obsn can be negative)
! rtrvz: retrieve z (obsn can be negative, dim(rz) = nfac + ncat + 1)
!
! Arguments:
!   obsn: observation indices
!   item: item index
!
! Return:
!   ra, rz: The value of a or z. If obsn > 0, they're simply the values from the
!   a and z matrices; for obsn < 0, they are values corresponding to the
!   artificial constraints

pure function rtrva(obsn, item) result(ra)
  integer, intent(in) :: obsn, item
  double precision :: ra

  ! artificial constraints
  if ( obsn .le. 0 ) then
    ! 0: upper bound for the 1st free intercept
    if (obsn .eq. 0) then
      ra = param_bnd
    ! -items(item)%nfr_int: lower bound for the last free intercept
    else if (obsn .eq. -items(item)%nfr_int) then
      ra = -param_bnd
    ! in between: ordering of intercepts
    else if (obsn .gt. -items(item)%nfr_int) then
      ra = 0.d0
    ! upper bound for free slopes
    else if ( mod(obsn + items(item)%nfr_int + 1, 2) .eq. 0 ) then
      ra = param_bnd
    ! lower bound for free slopes
    else
      ra = -param_bnd
    end if
  ! real people, left intercept
  else if ( obsn .le. nobsn ) then
    ra = a( obsn, item )
  ! real people, right intercept
  else
    ra = a( obsn - nobsn, item )
  end if
end function rtrva

pure function rtrvz(obsn, item) result(rz)
  integer, intent(in) :: obsn, item
  double precision :: rz( 0:items(item)%npar - 1 )

  rz = 0.d0
  ! artificial constraints
  if (obsn .le. 0) then
    ! 0: lower bound for the 1st free intercept
    if (obsn .eq. 0) then
      rz( items(item)%fr_int(1) ) = 1.d0
    ! -items(item)%nfr_int: upper bound for the last free intercept
    else if (obsn .eq. -items(item)%nfr_int) then
      rz( items(item)%fr_int(items(item)%nfr_int) ) = 1.d0
    ! in between: ordering of intercepts
    else if (obsn .gt. -items(item)%nfr_int) then
      rz( items(item)%fr_int(-obsn) ) = 1.d0
      rz( items(item)%fr_int(- obsn + 1) ) = -1.d0
    ! for free slopes
    else
      rz( items(item)%fr( items(item)%nfr_int + 1 - (obsn + items(item)%nfr_int + 1) / 2 ) ) = 1.d0
    end if
  ! real people, left intercept
  else if (obsn .le. nobsn) then
    rz( y(obsn, item) ) = 1.d0
    rz(items(item)%ncat + 1:items(item)%npar - 1) = z(obsn, :)
  ! real people, right intercept
  else
    rz( y(obsn - nobsn, item) + 1 ) = 1.d0
    rz(items(item)%ncat + 1:items(item)%npar - 1) = z(obsn - nobsn, :)
  end if
end function rtrvz

! sing: check singularity of a linear system
! solv: Solve for a linear system; the result gives a vertex. The number of
! linear equations should be the same as the length of obsn
!
! Arguments:
!   obsn: vector of observation indices
!   item: item index
!
! Output:
!   s: singular or not
!   v: a type(vnode) object

function sing(obsn, item) result(s)
  integer, intent(in) :: item, obsn(items(item)%nfr)
  logical :: s
  integer :: i
  double precision :: zobsn(1:items(item)%nfr, 0:items(item)%npar - 1), sc

  ! if all parameters fixed
  if ( items(item)%nfr .eq. 0 ) then
    s = .false.
    return
  end if
  
  ! retrieve z
  forall (i = 1:items(item)%nfr)
    zobsn(i, 0:items(item)%npar - 1) = rtrvz(obsn(i), item)
  end forall 

  ! check if LHS matrix is singular
  do i = 0, items(item)%nfr - 1
    sc = sum( zobsn(:, i) )
    if ( sc .eq. 0 ) then
      s = .true.
      return
    end if
  end do
  s = .false.

end function sing

subroutine solv(obsn, item, v)
  integer, intent(in) :: item, obsn(items(item)%nfr)
  type(vnode), intent(inout) :: v
  integer :: i, ipiv( items(item)%nfr ), info  ! LAPACK auxiliary variables
  double precision :: zobsn(1:items(item)%nfr, 0:items(item)%npar - 1), &
    lu( items(item)%nfr,items(item)%nfr )

  ! if all parameters fixed
  if ( items(item)%nfr .eq. 0 ) return

  ! retrieve a and z
  v%obsn = obsn
  forall (i = 1:items(item)%nfr)
    v%par(i) = rtrva(obsn(i), item)
    zobsn(i, 0:items(item)%npar - 1) = rtrvz(obsn(i), item)
  end forall 

  ! subtract fixed parts
  if ( items(item)%nfi .gt. 0 ) then
    v%par = v%par - matmul( zobsn(:, items(item)%fi), &
      items(item)%par( items(item)%fi ) )
  end if

  ! call LAPACK to solve
  lu = zobsn(:, items(item)%fr)
  call dgetrf(items(item)%nfr, items(item)%nfr, &
    lu, items(item)%nfr, ipiv, info)
  call dgetrs('N', items(item)%nfr, 1, lu, items(item)%nfr, ipiv, &
    v%par, items(item)%nfr, info)

end subroutine solv

! pushfront_v: Push v1 at the front vnode list v0
! pushback_v: Push v1 at the back vnode list v0
!
! Arguments:
!   v0: position
!   v1: new vnode
!
! Effect:
!   Check if v0 is associated; then push v1 next to v0

subroutine pushfront_v(v0, v1)
  type(vnode), pointer, intent(inout) :: v0, v1

  if ( .not. associated(v1) ) &
    stop "pushfront_v ERROR: v1 is null."
  if ( .not. associated(v0) ) then
    v0 => v1
    nullify(v0%prev, v0%next)
    return
  end if
  v0%prev => v1
  v1%next => v0
  nullify(v1%prev)
  v0 => v1
end subroutine pushfront_v

subroutine pushback_v(v0, v1)
  type(vnode), pointer, intent(inout) :: v0, v1

  if ( .not. associated(v1) ) &
    stop "pushback_v ERROR: v1 is null."
  if ( .not. associated(v0) ) then
    v0 => v1
    nullify(v0%prev, v0%next)
    return
  end if
  v0%next => v1
  v1%prev => v0
  nullify(v1%next)
  v0 => v1
end subroutine pushback_v

! rmvnode: Remove a vnode object from the linked list
!
! Arguments:
!   v: a vnode pointer
!   head: head pointer
!
! Effect:
!   Remove v from the linked-list. Deallocate its reference if it's not a
!   initial vertex. v is moved forward if possible.

subroutine rmvnode(v, head)
  type(vnode), pointer, intent(inout) :: v, head
  type(vnode), pointer :: rv
  logical :: prev, next

  rv => v
  prev = associated(v%prev)
  next = associated(v%next)
  ! others
  if (prev .and. next) then
    v%prev%next => v%next
    v%next%prev => v%prev
    v => v%next
  ! tail
  else if ( prev .and. ( .not. next ) ) then
    v%prev%next => v%next
    v => v%prev
  ! head
  else if ( ( .not. prev ) .and. next ) then
    v%next%prev => v%prev
    head => v%next
    v => v%next
  ! singleton
  else
    nullify(v, head)
  end if

  if (.not. rv%init) deallocate(rv)
end subroutine rmvnode

! reinitv: Re-initialize vlist by vinit
!
! Arguments: 
!   item: item index
!
! Effect:
!   Remove all existing elements in vlist; add in elements in vinit

subroutine reinitv(item)
  integer, intent(in) :: item
  type(vnode), pointer :: iv
  integer :: i

  ! clear vlist
  iv => items(item)%vlist
  do while ( associated(iv) )
    call rmvnode( iv, items(item)%vlist )
  end do
  ! copy elements from vinit
  items(item)%vlist => items(item)%vinit(1)%v
  iv => items(item)%vlist
  do i = 2, items(item)%nav
    items(item)%vinit(i)%v%prev => items(item)%vinit(i - 1)%v
    items(item)%vinit(i - 1)%v%next => items(item)%vinit(i)%v
  end do
  nullify( items(item)%vinit(items(item)%nav)%v%next )
end subroutine reinitv

! pushfront_e: Push e1 at the front of enode list e0
! pushback_e: Push e1 at the back of enode list e0
!
! Arguments:
!   e0: position
!   e1: new enode

subroutine pushfront_e(e0, e1)
  type(enode), pointer, intent(inout) :: e0, e1

  if ( .not. associated(e1) ) &
    stop "pushfront_e ERROR: e1 is null."
  if ( .not. associated(e0) ) then
    e0 => e1
    nullify(e0%prev, e0%next)
    return
  end if
  e0%prev => e1
  e1%next => e0
  nullify(e1%prev)
  e0 => e1
end subroutine pushfront_e

subroutine pushback_e(e0, e1)
  type(enode), pointer, intent(inout) :: e0, e1

  if ( .not. associated(e1) ) &
    stop "pushback_e ERROR: e1 is null."
  if ( .not. associated(e0) ) then
    e0 => e1
    nullify(e0%prev, e0%next)
    return
  end if
  e0%next => e1
  e1%prev => e0
  nullify(e1%next)
  e0 => e1
end subroutine pushback_e

! rmenode: Remove an enode object from the linked list
!
! Arguments:
!   e: a enode pointer
!   head: the head pointer
!
! Effect:
!   Remove e from the linked-list. Deallocate its reference if it's not a
!   initial vertex. e is moved forward if possible.

subroutine rmenode(e, head)
  type(enode), pointer, intent(inout) :: e, head
  type(enode), pointer :: re
  logical :: prev, next

  re => e
  prev = associated(e%prev)
  next = associated(e%next)
  ! others
  if (prev .and. next) then
    e%prev%next => e%next
    e%next%prev => e%prev
    e => e%next
  ! tail
  else if ( prev .and. ( .not. next ) ) then
    nullify(e%prev%next)
    e => e%prev
  ! head
  else if ( ( .not. prev ) .and. next ) then
    nullify(e%next%prev)
    head => e%next
    e => e%next
  ! singleton
  else
    nullify(e, head)
  end if
  if (.not. re%init) deallocate(re)
end subroutine rmenode

! reinite: Re-initialize elist by einit
!
! Arguments: 
!   item: item index
!
! Effect:
!   Remove all existing elements in elist; add in elements in einit

subroutine reinite(item)
  integer, intent(in) :: item
  type(enode), pointer :: ie
  integer :: i

  if (items(item)%nfr .eq. 0) return

  ! clear elist
  ie => items(item)%elist
  do while ( associated(ie) )
    call rmenode( ie, items(item)%elist )
  end do
  ! copy elements from einit
  items(item)%elist => items(item)%einit(1)%e
  ie => items(item)%elist
  do i = 2, items(item)%nae
    items(item)%einit(i)%e%prev => items(item)%einit(i - 1)%e
    items(item)%einit(i - 1)%e%next => items(item)%einit(i)%e
  end do
  nullify( items(item)%einit(items(item)%nae)%e%next )
end subroutine reinite

! pushfront_f: Push f1 at the front of fnode list f0
!
! Arguments:
!   f0: position
!   f1: new enode
!
! Effect:
!   Check if f0 is associated; then push f1 next to f0

!subroutine pushfront_f(f0, f1)
!  type(fnode), pointer, intent(inout) :: f0, f1
!
!  if ( .not. associated(f1) ) &
!    stop "pushfront_f ERROR: f1 is null."
!  if ( .not. associated(f0) ) then
!    f0 => f1
!    nullify(f0%prev, f0%next)
!    return
!  end if
!  f0%prev => f1
!  f1%next => f0
!  nullify(f1%prev)
!  f0 => f1
!end subroutine pushfront_f

! copy_ve: copy vlist v, and elist e
!
! Arguments:
!   v: a head node of vlist
!   e: a head node of elist
!   cv: a copy of v (cv is not allocated)
!   ce: a copy of e (ce is not allocated)
!
! Effect:
!   Copy v to cv, e to ce. The links between v and e are also replicated

!subroutine copy_ve(v, e, cv, ce)
!  type(vnode), pointer, intent(inout) :: v, cv
!  type(enode), pointer, intent(inout) :: e, ce
!  type(vnode), pointer :: iv, vu
!  type(enode), pointer :: ie, eu
!
!  ! copy v first, link v and cv with vcopy
!  allocate(cv)
!  iv => v
!  do while ( associated(iv) )
!    allocate(vu)
!    vu = iv
!    iv%vcopy => vu
!    call pushfront_v(cv, vu)
!    nullify(vu)
!    iv => iv%next
!  end do
!
!  ! copy e 
!  allocate(ce)
!  ie => e
!  do while ( associated(ie) )
!    allocate(eu)
!    eu = ie
!    ie%ecopy => eu
!    eu%v1 => ie%v1%vcopy
!    eu%v2 => ie%v2%vcopy
!    call pushfront_e(ce, eu)
!    nullify(eu)
!    ie => ie%next
!  end do
!
!end subroutine copy_ve


! rmfnode: Remove an fnode object from the linked list
!
! Arguments:
!   f: an fnode pointer
!   head: the head pointer
!
! Effect:
!   Remove f from the linked-list. Deallocate its reference if it's not a
!   initial vertex. f is moved forward if possible.

!subroutine rmfnode(f, head)
!  type(fnode), pointer, intent(inout) :: f, head
!  type(fnode), pointer :: ia
!  type(vnode), pointer :: iv
!  type(enode), pointer :: ie
!  logical :: prev, next
!
!  ! clear v_f
!  iv => f%v_f
!  do while ( associated(iv) )
!    call rmvnode( iv, f%v_f )
!  end do
!  ! clear e_f
!  ie => f%e_f
!  do while ( associated(ie) )
!    call rmenode( ie, f%e_f )
!  end do
!
!  ia => f
!  prev = associated(f%prev)
!  next = associated(f%next)
!  ! others
!  if (prev .and. next) then
!    f%prev%next => f%next
!    f%next%prev => f%prev
!    f => f%next
!  ! tail
!  else if ( prev .and. ( .not. next ) ) then
!    nullify(f%prev%next)
!    f => f%prev
!  ! head
!  else if ( ( .not. prev ) .and. next ) then
!    nullify(f%next%prev)
!    head => f%next
!    f => f%next
!  ! singleton
!  else
!    nullify(f, head)
!  end if
!  deallocate(ia)
!end subroutine rmfnode

! reinitef: Remove all nodes from items(item)%flist
!
! Arguments: 
!   item: item index
!
! Effect:
!   Remove all existing elements in flist

!subroutine reinitf(item)
!  integer, intent(in) :: item
!  type(fnode), pointer :: ia
!
!  if (items(item)%nfr .eq. 0) return
!
!  ! clear flist
!  ia => items(item)%flist
!  do while ( associated(ia) )
!    call rmfnode( ia, items(item)%flist )
!  end do
!end subroutine reinitf

end module test
